#include "analysis/analysis.h"

std::chrono::time_point<std::chrono::system_clock> start_chrono() {
    return std::chrono::system_clock::now();
}
std::chrono::time_point<std::chrono::system_clock> stop_chrono() {
    return std::chrono::system_clock::now();
}
int interval_chrono(std::chrono::time_point<std::chrono::system_clock> start, std::chrono::time_point<std::chrono::system_clock> stop) {
    return std::chrono::duration_cast<std::chrono::milliseconds>(stop-start).count();
}

double student_law_quartiles[21] = { 0.,
    12.7062047362, 4.3026527297, 3.1824463053, 2.7764451052, 2.5705818356,
    2.4469118511, 2.3646242516, 2.3060041352, 2.2621571628, 2.228138852,
    2.2009851601, 2.1788128297, 2.1603686565, 2.1447866879, 2.1314495456,
    2.1199052992, 2.1098155778, 2.1009220402, 2.0930240544, 2.0859634473
};

Statistic avg_calls(SearchHeuristic* algo, int nb, double abortion_limit, double student_law_quartile) {
    std::vector<double> data;
    Statistic stat;
    for(int i=0; i<nb; i++) {
        // Initialize algorithm
        if(abortion_limit > 0) {
            double lim = nb*abortion_limit-stat.sum + 1;
            if(lim <= 0)  lim = 1;
            algo->set_abortion_limit(lim);
        }
        // Execute algorithm on the problem
        Result res = algo->run();
        delete[] res.solution;
        int nb_calls = res.nb_calls;

        // Update statistics
        data.push_back(nb_calls);
        stat.sum += nb_calls;
        stat.deviation += pow(nb_calls, 2);

        if(i == 0 || nb_calls > stat.maximal)
            stat.maximal = nb_calls;
        if(i == 0 || nb_calls < stat.minimal)
            stat.minimal = nb_calls;

        // Verify if it was aborted
        if(res.has_been_aborted) {
            nb = i+1;
            break;
        }
    }

    // Calculate some statistics
    stat.nb = nb;
    stat.average = stat.sum / nb;
    stat.deviation = sqrt(stat.deviation/nb - pow(stat.average, 2));
    stat.no_bias_deviation = sqrt((double) nb/(nb-1)) * stat.deviation;
    if(student_law_quartile > 0)
        stat.alea = stat.no_bias_deviation * student_law_quartile / sqrt(nb);
    else
        stat.alea = (nb-1 <= 20) ? stat.no_bias_deviation * student_law_quartiles[nb-1] / sqrt(nb) : -1;
    return stat;
}

Statistic* avg_evolution(SearchHeuristic* algo, int max_evolution, int nb_tests) {
    Statistic* values = new Statistic[max_evolution];

    std::vector<Point> record;
    algo->set_record(&record);

    for(int i=0; i<nb_tests; i++) {
        Result res = algo->run();
        delete[] res.solution;

        Point p;
        int previous_nb_calls = -1;
        double previous_value = 0;
        for(int j=0, l=record.size(); j<=l; j++) {
            if(j < l)   p = record[j];
            else        p.nb_calls = max_evolution-1;

            int nb_calls = std::min(p.nb_calls, max_evolution-1);
            for(int k=previous_nb_calls+1; k<=nb_calls; k++) {
                if(k == nb_calls) {
                    previous_nb_calls = nb_calls;
                    previous_value = p.value;
                }
                values[k].nb++;
                values[k].sum += previous_value;
                values[k].deviation += pow(previous_value, 2);
            }
        }
    }
    for(int i=0; i<max_evolution; i++) {
        values[i].average = values[i].sum / values[i].nb;
        values[i].deviation = sqrt(values[i].deviation/values[i].nb - pow(values[i].average, 2));
        values[i].no_bias_deviation = sqrt((double) values[i].nb/(values[i].nb-1)) * values[i].deviation;
        values[i].alea = values[i].no_bias_deviation * 2. / sqrt(values[i].nb);
    }
    return values;
}

std::vector<Statistic> compute_complexity(std::default_random_engine& randomizer, SearchHeuristic* algo, int n_initial, int n_final, int n_step, int nb_tests, double slq, bool displaying) {
    std::vector<Statistic> stats;
    for(int n=n_initial; n<n_final; n+=n_step) {
        MasterMind mm(n, n, randomizer);
        algo->set_problem(&mm);
        Statistic stat = avg_calls(algo, nb_tests, -1, slq);
        stat.label = n;
        stats.push_back(stat);

        if(displaying)
            std::cout << n << ": " << stat.average << " +- " << stat.alea << std::endl;
    }
    return stats;
}

void save_in_file(std::string namefile, std::vector<Statistic> stats) {
    std::ofstream outfile(namefile);
    for(int i=0, s=stats.size(); i<s; i++)
        outfile << (int) stats[i].label << ";" << stats[i].average-stats[i].alea << ";" << stats[i].average+stats[i].alea << std::endl;
}

void cmp_evolution(std::vector<SearchHeuristic*> algos, MasterMind& problem, std::string outfilename, int max_evolution, int nb_tests) {
    int nb_algos = algos.size();
    Statistic** values = new Statistic*[nb_algos];
    for(int i=0; i<nb_algos; i++) {
        algos[i]->set_problem(&problem);
        values[i] = avg_evolution(algos[i], max_evolution, nb_tests);
    }
    std::ofstream statsfile(outfilename);
    for(int i=0; i<max_evolution; i++) {
        statsfile << i << ";";
        for(int k=0; k<nb_algos; k++)
            statsfile << values[k][i].average << ";";
        statsfile << std::endl;
    }

    for(int i=0; i<nb_algos; i++)
        delete[] values[i];
    delete[] values;
}
